
#define LED_BUILTIN 2

#include <FS.h>
#include <ArduinoJson.h>
#include <ESP8266WiFi.h>
#include <DNSServer.h>            //Local DNS Server used for redirecting all requests to the configuration portal
#include <ESP8266WebServer.h>     //Local WebServer used to serve the configuration portal
#include <WiFiManager.h>          //https://github.com/tzapu/WiFiManager WiFi Configuration Magic
#include <PubSubClient.h>         //https://github.com/knolleary/pubsubclient
WiFiClient espClient;
PubSubClient client(espClient);

#include <Ticker.h>
Ticker ticker;

long lastReconnectAttempt = 0;
bool shouldSaveConfig = false;
char mqtt_server[40] = "mqtt";
char mqtt_port[6] = "1883";
char mqtt_topic[40] = "sensors/sound-level";

void setup() {
	// Setup serial as 8E1 to communicate with sound meter
	Serial.begin(9600, SERIAL_8E1);
	Serial.setTimeout(100); // sound meter should respond relatively fast.

	// set led pin as output and blink 0.5s interval until connected
	pinMode(LED_BUILTIN, OUTPUT);
	ticker.attach(0.5, led_tick);
	
	load_settings();
	
	// WiFiManager
	// Local intialization. Once its business is done, there is no need to keep it around
	WiFiManager wifiManager;
	wifiManager.setAPCallback(configModeCallback);
	wifiManager.setSaveConfigCallback(saveConfigCallback);
	
	// custom parameters
	WiFiManagerParameter custom_mqtt_server("server", "mqtt server", mqtt_server, 40);
	wifiManager.addParameter(&custom_mqtt_server);
	WiFiManagerParameter custom_mqtt_port("port", "mqtt port", mqtt_port, 6);
	wifiManager.addParameter(&custom_mqtt_port);
	WiFiManagerParameter custom_mqtt_topic("topic", "mqtt topic", mqtt_topic, 40);
	wifiManager.addParameter(&custom_mqtt_topic);
		
	// try to connect or fallback to ESP+ChipID AP config mode.
	if ( ! wifiManager.autoConnect()) {
		// reset and try again, or maybe put it to deep sleep
		ESP.reset();
		delay(1000);
	}
	
	// read paramters
	strcpy(mqtt_server, custom_mqtt_server.getValue());
	strcpy(mqtt_port, custom_mqtt_port.getValue());
	strcpy(mqtt_topic, custom_mqtt_topic.getValue());
	
	// if we went through config, we should save our changes.
	if (shouldSaveConfig) save_settings();
	
	// no more blinking
	ticker.detach();
	
	// setup mqtt
	mqtt_connect();
	
	ticker.attach(1, process_data);
}


void loop() {
	if (!client.connected()) {
		long now = millis();
		
		if (now - lastReconnectAttempt > 5000) {
			lastReconnectAttempt = now;
			// Attempt to reconnect
			if (mqtt_reconnect()) {
				lastReconnectAttempt = 0;
			}
		}
	} else {
		// Client connected
		client.loop();
	}
}




void led_tick() {
	// toggle state
	int state = digitalRead(LED_BUILTIN);  // get the current state of GPIO1 pin
	digitalWrite(LED_BUILTIN, !state);     // set pin to the opposite state
}

//callback notifying us of the need to save config
void saveConfigCallback () {
	shouldSaveConfig = true;
}

// gets called when WiFiManager enters configuration mode
void configModeCallback(WiFiManager *myWiFiManager) {
	// entered config mode, make led toggle faster
	ticker.attach(0.2, led_tick);
}

void save_settings() {
	DynamicJsonBuffer jsonBuffer;
	JsonObject& json = jsonBuffer.createObject();
	json["mqtt_server"] = mqtt_server;
	json["mqtt_port"] = mqtt_port;
	json["mqtt_topic"] = mqtt_topic;

	File configFile = SPIFFS.open("/config.json", "w");
	// if ( ! configFile) {
		// Serial.println("failed to open config file for writing");
	// }
	
	json.printTo(configFile);
	configFile.close();
}

void load_settings() {

	if (SPIFFS.begin() && SPIFFS.exists("/config.json")) {
		File configFile = SPIFFS.open("/config.json", "r");
		
		if (configFile) {
			size_t size = configFile.size();
			// Allocate a buffer to store contents of the file.
			std::unique_ptr<char[]> buf(new char[size]);

			configFile.readBytes(buf.get(), size);
			DynamicJsonBuffer jsonBuffer;
			JsonObject& json = jsonBuffer.parseObject(buf.get());
			
			if (json.success()) {
				strcpy(mqtt_server, json["mqtt_server"]);
				strcpy(mqtt_port, json["mqtt_port"]);
				strcpy(mqtt_topic, json["mqtt_topic"]);
			}
		}
	}
}


void process_data() {
	digitalWrite(LED_BUILTIN, LOW);
	
	// write query
	Serial.write(0x30);
	Serial.write(0x01);
	Serial.write(0x0d);
	
	// read 4 bytes.
	// first 2 bytes are what we want.
	int a = Serial.read();
	int b = Serial.read();
	while (Serial.available() > 0) Serial.read();
	
	// Serial timed out or we got bad data
	if (a == -1 || b == -1) {
		// try to reset the connection in case it thinks we sent
		// an invalid command.
		Serial.write(0x0d);
		while (Serial.available() > 0) Serial.read();
		
		// initialize
		Serial.write(0x10);
		Serial.write(0x04);
		Serial.write(0x0d);
		delay(100);
		while (Serial.available() > 0) Serial.read();
		return;
	}
	
	// level value is found across 11 bits.
	int level = ((a & 0x07) << 8) | (b & 0xff);
	
	mqtt_push_value(level);
	
	digitalWrite(LED_BUILTIN, HIGH);
}



void mqtt_push_value(int x) {
	char msg[50];
	snprintf(msg, 50, "{\"value\": %ld}", x);
	
	client.publish(mqtt_topic, msg);
}

void mqtt_connect() {
	client.setServer(mqtt_server, atoi(mqtt_port));
	client.setCallback(mqtt_callback);
}

bool mqtt_reconnect() {
	// use a demo client name for now
	if (client.connect("ESPDemo")) {
		// send to system a reconnected message.
		client.publish("system", "MQTT connection established");
	}
	
	return client.connected();
}

void mqtt_callback(char* topic, byte* payload, unsigned int length) {
	// handle message arrived
}


